#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <getopt.h>
#include <stddef.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/spi/spidev.h>

#define CS_GPIO_PIN     (39)    //PAD_KEY0
#define CLK_GPIO_PIN    (40)    //PAD_KEY1
#define SDI_GPIO_PIN    (41)    //PAD_KEY2
#define SDO_GPIO_PIN    (42)    //PAD_KEY3

#define GPIO_HIGH       (1)
#define GPIO_LOW        (0)

#define SYSFS_GPIO_DIR  "/sys/class/gpio"
#define MAX_BUF         (64)

#define MSPI_DELAY      (2)

static int bits = 8;

/**
 * function: gpio export
 */
int ss_gpio_export(unsigned int gpio)
{
    int fd, len;
    char buf[MAX_BUF];

    fd = open(SYSFS_GPIO_DIR "/export", O_WRONLY);
    if (fd < 0) {
        perror("gpio/export");
        return fd;
    }

    len = snprintf(buf, sizeof(buf), "%d", gpio);
    write(fd, buf, len);
    close(fd);

    return 0;
}
 
/**
 * function: gpio unexport
 */
int ss_gpio_unexport(unsigned int gpio)
{
    int fd, len;
    char buf[MAX_BUF];

    fd = open(SYSFS_GPIO_DIR "/unexport", O_WRONLY);
    if (fd < 0) {
        perror("gpio/export");
        return fd;
    }

    len = snprintf(buf, sizeof(buf), "%d", gpio);
    write(fd, buf, len);
    close(fd);
    return 0;
}
 
/**
 * function:  gpio set direction
 * parameter: out_flag = 1 -- out
 *            out_flag = 0 -- in
 */
int ss_gpio_set_dir(unsigned int gpio, unsigned int out_flag)
{
    int fd;
    char buf[MAX_BUF];

    snprintf(buf, sizeof(buf), SYSFS_GPIO_DIR  "/gpio%d/direction", gpio);

    fd = open(buf, O_WRONLY);
    if (fd < 0) {
        perror("gpio/direction");
        return fd;
    }

    if (out_flag)
        write(fd, "out", 4);
    else
        write(fd, "in", 3);

    close(fd);
    return 0;
}
 
/**
 * function:  gpio set value
 * parameter: value = 1 -- high
 *            value = 0 -- low
 */
int ss_gpio_set_value(unsigned int gpio, unsigned int value)
{
    int fd;
    char buf[MAX_BUF];

    snprintf(buf, sizeof(buf), SYSFS_GPIO_DIR "/gpio%d/value", gpio);

    fd = open(buf, O_WRONLY);
    if (fd < 0) {
        perror("gpio/set-value");
        return fd;
    }

    if (value)
        write(fd, "1", 2);
    else
        write(fd, "0", 2);

    close(fd);
    return 0;
}
 
/**
 * function: gpio get value
 */
int ss_gpio_get_value(unsigned int gpio, unsigned int *value)
{
    int fd;
    char buf[MAX_BUF];
    char ch;

    snprintf(buf, sizeof(buf), SYSFS_GPIO_DIR "/gpio%d/value", gpio);

    fd = open(buf, O_RDONLY);
    if (fd < 0) {
        perror("gpio/get-value");
        return fd;
    }

    read(fd, &ch, 1);

    if (ch != '0') {
        *value = 1;
    } else {
        *value = 0;
    }

    close(fd);
    return 0;
}
 
 
/**
 * function:  gpio_set edge
 * parameter: edge = "none", "rising", "falling", "both"
 */
int ss_gpio_set_edge(unsigned int gpio, char *edge)
{
    int fd;
    char buf[MAX_BUF];
 
    snprintf(buf, sizeof(buf), SYSFS_GPIO_DIR "/gpio%d/edge", gpio);

    fd = open(buf, O_WRONLY);
    if (fd < 0) {
        perror("gpio/set-edge");
        return fd;
    }

    write(fd, edge, strlen(edge) + 1);
    close(fd);
    return 0;
}
 
/**
 * function: gpio open fd
 */
int ss_gpio_open(unsigned int gpio)
{
    int fd;
    char buf[MAX_BUF];

    snprintf(buf, sizeof(buf), SYSFS_GPIO_DIR "/gpio%d/value", gpio);

    fd = open(buf, O_RDONLY | O_NONBLOCK );
    if (fd < 0) {
        perror("gpio/fd_open");
    }
    return fd;
}
 
/**
 * function: gpio close fd
 */
int ss_gpio_close(int fd)
{
    return close(fd);
}

/**
 * function: delay for some time
 */
void ss_cmd_delay(unsigned int i)
{
    unsigned int j, cnt;
    cnt = i * 2;
    for (j = 0; j < cnt; j++);
}

/**
 * function: init gpio for spi
 */
int ss_mspi_init(void)
{
    ss_gpio_export(CS_GPIO_PIN);
    ss_gpio_export(CLK_GPIO_PIN);
    ss_gpio_export(SDI_GPIO_PIN);
    ss_gpio_export(SDO_GPIO_PIN);

    ss_gpio_set_dir(CS_GPIO_PIN, 1);
    ss_gpio_set_dir(CLK_GPIO_PIN, 1);
    ss_gpio_set_dir(SDI_GPIO_PIN, 1);
    ss_gpio_set_dir(SDO_GPIO_PIN, 0);

    //simulating mspi mode 0: CPOL=0, CPHA=0
    //cs keep high in idle
    ss_gpio_set_value(CS_GPIO_PIN, GPIO_HIGH);
    //clk keep low in idle
    ss_gpio_set_value(CLK_GPIO_PIN, GPIO_LOW);
    ss_gpio_set_value(SDI_GPIO_PIN, GPIO_LOW);

    return 0;
}

/**
 * function: mspi write and read data, msb first
 * parameter: value -- u16 data
 *            bits  -- transmit bits
 */
uint16_t ss_mspi_write_read(uint16_t data, int bits)
{
    int i;
    unsigned int bit;
    uint16_t tmp = 0;

    ss_cmd_delay(MSPI_DELAY);

    for(i = 0; i < bits; i += 1)
    {
        ss_cmd_delay(MSPI_DELAY);

        //write bit
        if(data & (0x01 << (bits - 1)))
            ss_gpio_set_value(SDI_GPIO_PIN, GPIO_HIGH);
        else
            ss_gpio_set_value(SDI_GPIO_PIN, GPIO_LOW);
        data <<= 1;

        ss_gpio_set_value(CLK_GPIO_PIN, GPIO_LOW);
        ss_cmd_delay(MSPI_DELAY);
        ss_gpio_set_value(CLK_GPIO_PIN, GPIO_HIGH);

        //read bit
        ss_gpio_get_value(SDO_GPIO_PIN, &bit);
        tmp += bit;
        tmp <<= ((i < bits - 1) ? 1 : 0);
    }

    ss_cmd_delay(MSPI_DELAY);
    ss_gpio_set_value(SDI_GPIO_PIN, GPIO_LOW);
    ss_gpio_set_value(CLK_GPIO_PIN, GPIO_LOW);

    return tmp;
}

/**
 * function: mspi write and read test
 */
int ss_mspi_lookback(uint8_t *wr_buf, uint8_t *rd_buf, int size)
{
    int i;
    uint16_t wr_data, rd_data;

    if (bits > 8)
    {
        for (i = 0; i < size && size > 1; i += 2)
        {
            wr_data = (wr_buf[i + 1] << 8) | wr_buf[i];
            rd_data = ss_mspi_write_read(wr_data, bits);
            rd_buf[i]     = rd_data;
            rd_buf[i + 1] = rd_data >> 8;
        }
    }
    else
    {
        for (i = 0; i < size; i += 1)
        {
            wr_data = wr_buf[i];
            rd_data = ss_mspi_write_read(wr_data, bits);
            rd_buf[i] = rd_data;
        }
    }

    return 0;
}

int main(int argc, char *argv[])
{
    int i, len;
    uint8_t tmp_buf[64];
    uint8_t wr_buf[64], rd_buf[64];

    if (argc < 3)
    {
        printf("please inpu such as: ./mspi_gpio bits string\n");
        printf("eg: ./mspi_gpio 8 FF\n");
        printf("eg: ./mspi_gpio 9 01FF\n");
        return -1;
    }

    printf("mspi test start!\n");

    //set mspi transmit bits
    bits = atoi(argv[1]);

    ss_mspi_init();

    memset(wr_buf, 0x0, sizeof(wr_buf));
    memset(rd_buf, 0x0, sizeof(rd_buf));
    memset(tmp_buf, 0x0, sizeof(rd_buf));

    //receive shell '0'~'F' string and convert to char data
    uint8_t *cp = argv[2];
    uint8_t tmp = 0;
    for (i = 0; *cp; i ++, cp ++)
    {
        tmp = *cp - '0';
        if (tmp > 9)
            tmp -= ('A' - '0') - 10;
        if (tmp > 15)
            tmp -= ('a' - 'A');
        if (tmp > 15)
        {
            printf("invalid hex data on %c!\n", *cp);
            return -1;
        }

        if (i % 2)
            tmp_buf[i / 2] |= tmp;
        else
            tmp_buf[i / 2]  = tmp << 4;
    }
    len = i / 2;
    printf("test len: %d, bits: %d, in data: %s\n", len, bits, argv[2]);

    //save data to an array
    if (bits <= 8)
    {
        //save data to uint8_t
        for (i = 0; i < len; i ++)
        {
            wr_buf[i] = tmp_buf[i];
        }
    }
    else
    {
        //need to convert to uint16_t, high bits save to high address
        for (i = 0; i < len; i += 2)
        {
            wr_buf[i + 1] = tmp_buf[i];
            wr_buf[i] = tmp_buf[i + 1];
        }
    }

    for (i = 0; i < len; i ++)
    {
        if (!(i % 8))
            printf("\ninput  data: ");
        printf("0x%02x ", wr_buf[i]);
    }
    printf("\n\n");

    //set cs low, enable spi device
    ss_gpio_set_value(CS_GPIO_PIN, GPIO_LOW);

    ss_mspi_lookback(wr_buf, rd_buf, len);

    ss_gpio_set_value(CS_GPIO_PIN, GPIO_HIGH);

    for (i = 0; i < len; i ++)
    {
        if (!(i % 8))
            printf("\noutput data: ");
        printf("0x%02x ", rd_buf[i]);
    }
    printf("\n");

    return 0;
}
